/***********************************************************

    hbs reduction subroutine: hbscbin.c

                                   1998/06/10  K. Kawabata
                                   1998/10/05  K. Kawabata

    Count binning

    usage: % hbscbin infn reffile photfn outfn mode modified_reffile

      infn -- input spectrum (FITS format)
      reffile -- binning table reference
      photfn -- photometric parameter file (hbsphfit3)
      outfn -- output spectrum (FITS format)
      mode -- 1: sum (1 weight)  2: average (1 weight)  3: average (2 weights)
      modified_reffile -- effective wavelength overwrited

************************************************************/

#include<stdio.h>
#include<string.h>
#include<math.h>
#include<stdlib.h>
#include<unistd.h>
#include"fitsio.h"
#include"hbsinterpol.c"

#include"hbsdefine.h"

/*
const char *FILH = "fil_";
const char *FILT = ".dat";
*/

int scanfile( char *infile, float count[]  ){

  fitsfile *fp;
  int i, anynul, status;
  char buf[100];
  float nulval = 0.;
  long naxis1, naxis2, npix;
  char comment[100];

  status = 0;

  if( fits_open_file( &fp, infile, READONLY, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS1", &naxis1,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS2", &naxis2,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( naxis1 != NX2 || naxis2 != 1 ){
    printf( "error: definition of NX2 or NY(=1) in 'hbsdefine.h'\n" );
    return( -1 );
  }
  npix = naxis1 * naxis2;

  if( fits_read_img( fp, TFLOAT, 1L, npix, &nulval,\
      count, &anynul, &status ) != 0 ){
    return( status );
  }
  if( fits_close_file( fp, &status ) != 0 ){
    return( status );
  }
  return( 0 );
}

int readheadi( char *infile, char *keyword  ){

  fitsfile *fpf;
  int status;
  int value=0;
  char comment[100];

  status = 0;

  if( fits_open_file( &fpf, infile, READONLY, &status ) != 0 ){
    return( -1 );
  }
  if( fits_read_key( fpf, TINT, keyword, &value,\
      comment, &status ) != 0 ){
    return( -1 );
  }
  if( fits_close_file( fpf, &status ) != 0 ){
    return( status );
  }
  return( value );
}

/* In case of non-weighting */

double realcount3( float c[], double x_low, double x_high, double *wlmd, double *ewid ){

  int ix_low, ix_high;
  int i, j;
  double count, tcount = 0.0;
  int rswp, rewp;

  *wlmd = 0.0;
  *ewid = 0.0;

  rswp = ( RSW - SW + DW / 2.0 + 0.00000001 ) / DW + 1.0 ;
  rewp = ( REW - SW + DW / 2.0 + 0.00000001 ) / DW + 1.0 ;

  if( x_low >= x_high || ( x_low < ( rswp - 0.5 ) && x_high < ( rswp - 0.5 ) )\
     || ( x_low > ( rewp + 0.5 ) && x_high > ( rewp + 0.5 ) ) ){
    tcount = 0.0;
  }
  else{
    if( x_low < ( rswp - 0.5 ) )
      x_low = rswp - 0.5;
    if( x_high > ( rewp + 0.5 ) )
      x_high = rewp + 0.5;

    ix_low = x_low + 1.4999999999;
    ix_high = x_high - 0.4999999999;

    if( ix_low <= ix_high ){
      for( j=ix_low; j<=ix_high; j++ ){
        count = c[j-1];
	tcount += count;
	*wlmd += ( count * j );
      }
    }
    if( ( ix_high - ix_low ) > -2 ){
      if( (ix_low - 1) >= rswp ){
        count = ( ( ( ix_low - 0.5 ) - x_low ) * c[ix_low-2] );
	tcount += count;
	*wlmd += ( count * ( ix_low - 1 ) );
      }
      if( ix_high <= ( rewp - 1 ) ){
        count = ( ( x_high - ( ix_high + 0.5 ) ) * c[ix_high] );
	tcount += count;
	*wlmd += ( count * ( ix_high + 1 ) );
      }
    }
    else{
      ix_high = ix_high + 1;
      ix_low = ix_low - 1;
      if( ix_high != ix_low ){
        printf( "Error in realcount\n" );
        return( 0.0 );
      }
      count = ( ( x_high - x_low ) * c[ix_high-1] );
      tcount += count;
      *wlmd += ( count * ix_high );
    }
  }
  if( tcount != 0.0 )
    *wlmd = ( *wlmd / tcount - 1.0 ) * DW + SW;
  i = ( *wlmd - SW + DW / 2.0 ) / DW;
  if( fabs( c[i] ) > 1.0e-100 )
    *ewid = ( tcount / c[i] ) * DW;
  return( tcount );
}

/* In case of weighting by filter */

double realcount4( float c[], double wl[], double bw[], int n, double lsr[], double sr[], int nsr, double *wlmd, double *ewid, double *intwid ){

  int ix_low, ix_high;
  int i, j;
  double a, lmd, sysres;
  double count, tcount = 0.0;
  double comwei;

  *wlmd = 0.0;
  *ewid = 0.0;
  *intwid = 0.0;
  for( i=1; i<=NX2; i++ ){
    lmd = SW + ( i - 1.0 ) * DW;
    if( lmd >= RSW && lmd <= REW ){
      a = interpol( n, wl, bw, lmd, 1 );
      sysres = interpol( nsr, lsr, sr, lmd, 1 );
      comwei = a * sysres;
      count = c[i-1] * comwei;
      tcount += count;
      *wlmd += ( count * lmd );
      *intwid += comwei;
/* printf( "%4d  %7.2f %6.4f %e %e\n", i, lmd, a, count, tcount ); */
    }
  }
  *wlmd /= tcount;
  i = ( *wlmd - SW + DW / 2.0 ) / DW;
  if( c[i] > 1.0e-30 ){
    a = interpol( n, wl, bw, *wlmd, 1 );
    sysres = interpol( nsr, lsr, sr, *wlmd, 1 );
    *ewid = ( tcount / ( c[i] * a * sysres ) ) * DW;
  }
  return( tcount );
}

/* In case of weighting by filter & ? */

double realcount5( float c[], double x_low, double x_high, double wl[], double bw[], int n, double lsr[], double sr[], int nsr, double *wlmd, double *ewid ){

  int ix_low, ix_high;
  int i, j;
  double count, tcount = 0.0;
  int rswp, rewp;
  double a, totweit = 0.0;
  double lmd, sysres;
  double comwei;

  *wlmd = 0.0;
  *ewid = 0.0;

  rswp = ( RSW - SW + DW / 2.0 + 0.000000001 ) / DW + 1.0 ;
  rewp = ( REW - SW + DW / 2.0 + 0.000000001 ) / DW + 1.0 ;

  if( x_low >= x_high || ( x_low < ( rswp - 0.5 ) && x_high < ( rswp - 0.5 ) ) || ( x_low > ( rewp + 0.5 ) && x_high > ( rewp + 0.5 ) ) ){
    tcount = 0.0;
  }
  else{
    if( x_low < ( rswp - 0.5 ) )
      x_low = rswp - 0.5;
    if( x_high > ( rewp + 0.5 ) )
      x_high = rewp + 0.5;

    ix_low = x_low + 1.4999999999;
    ix_high = x_high - 0.4999999999;

    if( ix_low <= ix_high ){
      for( j=ix_low; j<=ix_high; j++ ){
        lmd = ( j - 1.0 ) * DW + SW;
	a = interpol( n, wl, bw, lmd, 1 );
        sysres = interpol( nsr, lsr, sr, lmd, 1 );
	comwei = a * sysres;
        count = c[j-1] * comwei;
	tcount += count;
	totweit += comwei;
	*wlmd += ( count * j );
      }
    }
    if( ( ix_high - ix_low ) > -2 ){
      if( (ix_low - 1) >= rswp ){
        lmd = ( ix_low - 1.0 ) * DW + SW;
	a = interpol( n, wl, bw, lmd, 1 );
        sysres = interpol( nsr, lsr, sr, lmd, 1 );
	comwei = a * sysres * ( ( ix_low - 0.5 ) - x_low );
        count = c[ix_low-2] * comwei;
	tcount += count;
	totweit += comwei;
	*wlmd += ( count * ( ix_low - 1 ) );
      }
      if( ix_high <= ( rewp - 1 ) ){
        lmd = ( ix_high - 1.0 ) * DW + SW;
	a = interpol( n, wl, bw, lmd, 1 );
        sysres = interpol( nsr, lsr, sr, lmd, 1 );
	comwei = a * sysres * ( x_high - ( ix_high + 0.5 ) );
        count = c[ix_high] * comwei;
	tcount += count;
	totweit += comwei;
	*wlmd += ( count * ( ix_high + 1 ) );
      }
    }
    else{
      ix_high = ix_high + 1;
      ix_low = ix_low - 1;
      if( ix_high != ix_low ){
        printf( "Error in realcount\n" );
        return( 0.0 );
      }
      lmd = ( ix_high - 1.0 ) * DW + SW;
      a = interpol( n, wl, bw, lmd, 1 );
      sysres = interpol( nsr, lsr, sr, lmd, 1 );
      comwei = a * sysres * ( x_high - x_low );
      count = c[ix_high-1] * comwei;
      tcount += count;
      totweit += comwei;
      *wlmd += ( count * ix_high );
    }
  }
  if( totweit != 0.0 )
    tcount /= totweit;
  if( tcount != 0.0 )
    *wlmd = ( *wlmd / tcount - 1.0 ) * DW + SW;
  i = ( *wlmd - SW + DW / 2.0 ) / DW;
  if( fabs( c[i] ) > 1.0e-100 ){
    a = interpol( n, wl, bw, *wlmd, 1 );
    sysres = interpol( nsr, lsr, sr, *wlmd, 1 );
    *ewid = ( tcount / ( c[i] * a * sysres ) ) * DW;
  }
  return( tcount );
}

/* In case of weighting by filter and cout */

double realcount6( float c[], double wl1[], double bw1[], int n1, double wl2[], double bw2[], int n2, double lsr[], double sr[], int nsr, double *wlmd, double *ewid ){

  int ix_low, ix_high;
  int i, j;
  double a1, a2, lmd;
  double count, tcount = 0.0;
  double sysres, intwid = 0.0;
  double comwei;

  *wlmd = 0.0;
  *ewid = 0.0;
  for( i=1; i<=NX2; i++ ){
    lmd = SW + ( i - 1.0 ) * DW;
    if( lmd >= RSW && lmd <= REW ){
      a1 = interpol( n1, wl1, bw1, lmd, 1 );
      a2 = interpol( n2, wl2, bw2, lmd, 1 );
      sysres = interpol( nsr, lsr, sr, lmd, 1 );
      comwei = a1 * a2 * sysres;
      count = c[i-1] * comwei;
      tcount += count;
      *wlmd += ( count * lmd );
      intwid += comwei;
    }
  }
  *wlmd /= tcount;
  i = ( *wlmd - SW + DW / 2.0 ) / DW;
  if( c[i] > 1.0e-30 ){
    a1 = interpol( n1, wl1, bw1, *wlmd, 1 );
    a2 = interpol( n2, wl2, bw2, *wlmd, 1 );
    sysres = interpol( nsr, lsr, sr, *wlmd, 1 );
    *ewid = ( tcount / ( c[i] * a1 * a2 * sysres ) ) * DW;
  }
  if( intwid != 0.0 )
    tcount /= intwid;
  else
    tcount = -9.9e+30;
  return( tcount );
}

int writefile( char *outfile, float count[] ){

  fitsfile *fp;
  int i, status;
  char buf[100];
  long naxis1, naxis2, npix;
  char comment[100];

  status = 0;

  if( fits_open_file( &fp, outfile, READWRITE, &status ) != 0 ){
    printf( "Cannot open file '%s' with READWRITE mode. status=%d\n",\
            outfile, status );
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS1", &naxis1,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( fits_read_key( fp, TLONG, "NAXIS2", &naxis2,\
      comment, &status ) != 0 ){
    return( status );
  }
  if( naxis1 != NX2 ){
    printf( "Error: definition of NX2 in 'hbsdefine.h'\n" );
    return( -1 );
  }
  if( naxis2 != 1 ){
    printf( "Error: definition of y-dimension in 'hbsdefine.h'\n" );
    return( -1 );
  }
  npix = naxis1 * naxis2;

  if( fits_write_img( fp, TFLOAT, 1, npix,\
      count, &status ) != 0 ){
    printf( "Error in the function fits_write_img\n" );
    return( status );
  }
  if( fits_close_file( fp, &status ) != 0 ){
    return( status );
  }

  return( 0 );
}

int main( int argc, char *argv[] ){

  float c1[NX2], c2[NX2];

/*
  char filsymb[][30] = { "JU", "JB", "JV", "CR", "CI" };
  char filfile[][30] = { "fil_JU.dat", "fil_JB.dat", "fil_JV.dat", "fil_CR.dat", "fil_CI.dat" };
  int nfil = 5;
*/
  char fn_filter[50], refdir[30];
  double fil_wl[50], fil_bp[50];
  int fil_n;
  char *okfile = "hbs.tmp.hbscbin";

  char reffile[50];
  char infile[50];
  char outfile[50];
  char refout[50];
  char photfile[50];

  char weight2[50]; /* weight file ..  assuming *i2.xy */
  double xweit[NX2], yweit[NX2];
  int nweit;
  int pregain, postgain;

  int nphot;
  double xphot[NX2], yphot[NX2];

  double binc[NX2]; /* binning center lambda(A) */
  double tmp_binc; /* effective binc (count-weighted-mean) */
  char binf[NX2][15]; /* ascii characters of binning width(A) */
  double binw; /* binning width(A) */
  double tmp_binw; /* effective binw */
  double a, integw;
  int nbin;
  double xbinc, xbinw;
  int mode; /* 1: sum  2: bin */

  FILE *fp, *fp2;
  char buf[200], last_buf[200];
  int i, j, flag_fil;
  int flag_refout = 0;

/***** confirmation of input file *****/

  if( argc != 6  && argc != 7 && argc != 8 ){
    printf( "usage: hbscbin infn ref_fn ref_dir phot_fn outfn mode [second-weight]  [modified_ref]\n" );
    exit( 1 );
  }

  strcpy( infile, argv[1] );
  strcpy( reffile, argv[2] );
  strcpy( refdir, argv[3] );
  strcpy( photfile, argv[4] );
  strcpy( outfile, argv[5] );
  mode = atoi( argv[6] );

  if( argc >= 8 ){
    if( mode != 3 ){
      strcpy( refout, argv[7] );
      flag_refout = 1;
    }
    else{
      strcpy( weight2, argv[7] );
      if( argc == 9 ){
	strcpy( refout, argv[8] );
	flag_refout = 1;
      }
    }
  }

  /*** delete old OK-file ***/

  if( access( okfile, 0 ) == 0 )
    remove( okfile );

  if( mode != 1 && mode != 2 && mode != 3 ){
    printf( "Error in mode number '%d'\n", mode );
    exit( -1 );
  }

  for( i=0; i<=(NX2-1); i+=1 ){
    c2[i] = 0.0;
  }

/* read input file */

  if( scanfile( infile, c1 ) != 0 ){
    printf( "Read error in '%s'\n", infile );
    exit( 1 );
  }

  if( mode == 3 ){

/* read second weight (count) file */

    if( ( fp = fopen( weight2, "rt" ) ) == NULL ){
      printf( "Cannot open '%s'\n", weight2 );
      exit( 1 );
    }
    for( i=0; i<=(NX2-1); i+=1 ){
      fgets( buf, 200, fp );
      sscanf( buf, "%lf %lf", &xweit[i], &yweit[i] );
    }
    nweit = NX2 - 1;
  }

/* read photometric parameter file */

  if( strcmp( photfile, "NONE" ) != 0 ){

    pregain = readheadi( infile, "PREGAIN" );
    if( pregain == -1 ){
      printf( "Cannot find paramater 'PREGAIN' in %s\n", infile );
      exit( -1 );
    }
    postgain = readheadi( infile, "POSTGAIN" );
    if( postgain == -1 ){
      printf( "Cannot find paramater 'POSTGAIN' in %s\n", infile );
      exit( -1 );
    }

    nphot = -1;
    if( ( fp = fopen( photfile, "rt" ) ) == NULL ){
      printf( "Photmetric parameter file '%s' not found.\n", photfile );
      exit( -1 );
    }
    while( feof( fp ) == 0 ){
      fgets( buf, 200, fp );
      if( strcmp( buf, last_buf ) == 0 )
	continue;
      nphot++;
      if( nphot >= NX2 ){
	printf( "n_phot exceeds NX2.\n" );
	break;
      }
      sscanf( buf, "%lf %lf %lf %lf %lf %lf", &xphot[nphot], &a, &a, &a, &yphot[nphot], &a );
      yphot[nphot] *= ( GEADU * pregain * postgain * 2.9979e+26 * 1.0e4 / ( xphot[nphot] * xphot[nphot] ) );
      strcpy( last_buf, buf );
    }
    fclose(fp);
  }
  else{
    nphot = 1;
    xphot[0] = 0.000;
    xphot[1] = 20000.000;
    yphot[0] = 1.000;
    yphot[1] = 1.000;
  }

/* read binning-factor reference file */

  nbin = -1;
  if( ( fp = fopen( reffile, "rt" ) ) == NULL ){
    printf( "Binning-table '%s' not found.\n", reffile );
    exit( -1 );
  }
  while( feof( fp ) == 0 ){
    fgets( buf, 200, fp );
    if( strcmp( buf, last_buf ) == 0 )
      continue;
    nbin++;
    if( nbin >= NX2 ){
      printf( "n_bin exceeds NX2.\n" );
      break;
    }
    sscanf( buf, "%lf %s", &binc[nbin], binf[nbin] );
    strcpy( last_buf, buf );
  }
  fclose(fp);

/*** binning ***/

  if( flag_refout == 1 ){
    if( ( fp2 = fopen( refout, "wt" ) ) == NULL ){
      printf( "Binning-table '%s' cannot open with write-mode.\n", refout );
      exit( -1 );
    }
  }
  if( mode == 1 ){
    for( i=0; i<=nbin; i+=1 ){
      flag_fil = 0;
      if( atoi( binf[i] ) == 0 ){
        sprintf( fn_filter, "%s%s%s%s", refdir, FILH, binf[i], FILT );
	flag_fil = 1;
      }
      if( flag_fil == 0 ){
	xbinc = ( binc[i] - SW ) / DW + 1.0;
	xbinw = atof( binf[i] ) / DW;
	c2[i] = realcount3( c1, (xbinc-0.5*xbinw), (xbinc+0.5*xbinw), &tmp_binc, &tmp_binw );
      }
      else{
	if( ( fp = fopen( fn_filter, "rt" ) ) == NULL ){
	  printf( "Filter passband file '%s' not found.\n", fn_filter );
	  exit( -1 );
	}
	fil_n = -1;
	while( feof( fp ) == 0 ){
	  fgets( buf, 200, fp );
	  if( strcmp( buf, last_buf ) == 0 )
	    continue;
	  fil_n++;
	  sscanf( buf, "%lf %lf", &fil_wl[fil_n], &fil_bp[fil_n] );
	  strcpy( last_buf, buf );
	}

	fclose( fp );
	if( fil_n == -1 ){
	  printf( "Error in reading '%s'\n", fn_filter );
	  exit( -1 );
	}
	a = realcount4( c1, fil_wl, fil_bp, fil_n, xphot, yphot, nphot, &tmp_binc, &tmp_binw, &integw );
	c2[i] = a;
      }
      if( flag_refout == 1 ){
	if( fabs( tmp_binc ) > 1.0e-30 );
	else
	  tmp_binc = binc[i];
	if( fabs( tmp_binw ) > 1.0e-30 );
	else
	  tmp_binw = atof( binf[i] );
	fprintf( fp2, "%9.4f %s  %9.4f %9.4f\n", binc[i], binf[i], tmp_binc, tmp_binw );
      }
    }
  }
  if( mode == 2 ){
    for( i=0; i<=nbin; i+=1 ){
      flag_fil = 0;
      if( atoi( binf[i] ) == 0 ){
        sprintf( fn_filter, "%s%s%s%s", refdir, FILH, binf[i], FILT );
/*	  sprintf( fn_filter, "%s%s", refdir, filfile[j] ); */
	flag_fil = 1;
      }
      if( flag_fil == 0 ){
	xbinc = ( binc[i] - SW ) / DW + 1.0;
	xbinw = atof( binf[i] ) / DW;
	c2[i] = ( realcount3( c1, (xbinc-0.5*xbinw), (xbinc+0.5*xbinw), &tmp_binc, &tmp_binw ) / xbinw );
      }
      else{
	if( ( fp = fopen( fn_filter, "rt" ) ) == NULL ){
	  printf( "Filter passband file '%s' not found.\n", fn_filter );
	  exit( -1 );
	}
	fil_n = -1;
	while( feof( fp ) == 0 ){
	  fgets( buf, 200, fp );
	  if( strcmp( buf, last_buf ) == 0 )
	    continue;
	  fil_n++;
	  sscanf( buf, "%lf %lf", &fil_wl[fil_n], &fil_bp[fil_n] );
	  strcpy( last_buf, buf );
	}

	fclose( fp );
	if( fil_n == -1 ){
	  printf( "Error in reading '%s'\n", fn_filter );
	  exit( -1 );
	}
	a = realcount4( c1, fil_wl, fil_bp, fil_n, xphot, yphot, nphot, &tmp_binc, &tmp_binw, &integw );
	c2[i] = a / integw;
      }
      if( flag_refout == 1 ){
	if( fabs( tmp_binc ) > 1.0e-30 );
	else
	  tmp_binc = binc[i];
	if( fabs( tmp_binw ) > 1.0e-30 );
	else
	  tmp_binw = atof( binf[i] );
	fprintf( fp2, "%9.4f %9.4f  %9.4f %s\n", tmp_binc, tmp_binw, binc[i], binf[i] );
      }
    }
  }
  if( mode == 3 ){
    for( i=0; i<=nbin; i+=1 ){
      flag_fil = 0;
      if( atoi( binf[i] ) == 0 ){
        sprintf( fn_filter, "%s%s%s%s", refdir, FILH, binf[i], FILT );
	/*	  sprintf( fn_filter, "%s%s", refdir, filfile[j] );*/
	flag_fil = 1;
      }
      if( flag_fil == 0 ){
	xbinc = ( binc[i] - SW ) / DW + 1.0;
	xbinw = atof( binf[i] ) / DW;
	c2[i] = ( realcount5( c1, (xbinc-0.5*xbinw), (xbinc+0.5*xbinw), xweit, yweit, nweit, xphot, yphot, nphot, &tmp_binc, &tmp_binw ) / xbinw );
      }
      else{
	if( ( fp = fopen( fn_filter, "rt" ) ) == NULL ){
	  printf( "Filter passband file '%s' not found.\n", fn_filter );
	  exit( -1 );
	}
	fil_n = -1;
	while( feof( fp ) == 0 ){
	  fgets( buf, 200, fp );
	  if( strcmp( buf, last_buf ) == 0 )
	    continue;
	  fil_n++;
	  sscanf( buf, "%lf %lf", &fil_wl[fil_n], &fil_bp[fil_n] );
	  strcpy( last_buf, buf );
	}

	fclose( fp );
	if( fil_n == -1 ){
	  printf( "Error in reading '%s'\n", fn_filter );
	  exit( -1 );
	}
	c2[i] = realcount6( c1, fil_wl, fil_bp, fil_n, xweit, yweit, nweit, xphot, yphot, nphot, &tmp_binc, &tmp_binw );
      }
      if( flag_refout == 1 ){
	if( fabs( tmp_binc ) > 1.0e-30 );
	else
	  tmp_binc = binc[i];
	if( fabs( tmp_binw ) > 1.0e-30 );
	else
	  tmp_binw = atof( binf[i] );
	fprintf( fp2, "%9.4f %9.4f  %9.4f %s\n", tmp_binc, tmp_binw, binc[i], binf[i] );
      }
    }
  }
  if( flag_refout == 1 )
    fclose( fp2 );

/* write output file */

  if( writefile( outfile, c2 ) != 0 ){
    printf( "write failed in '%s'\n", outfile );
    return( 1 );
  }

/* write OK-file */
  if( ( fp = fopen( okfile, "wt" ) ) == NULL ){
    printf( "Cannot open OK-file '%s'\n", okfile );
    exit( -1 );
  }
  fprintf( fp, "OK\n" );
  fclose( fp );

  return( 0 );
}
